r"""Compute time and memory consumption.

Linting
-------
```sh
$ pylint src/
```

Usage
-----
You can go on:
https://github.com/FilouPlains/perfassess

To check how to install this module. After, you can launch the module as a
command line. To know what is possible, launch:
```sh
$ perfassess --help
```

Else, you are also able to import the module object as followed:

```py
from perfassess.class_performance_assessor import PerformanceAssessor
from perfassess.testor import testor

assessor: PerformanceAssessor = PerformanceAssessor(
    main=testor,
    n_field=1,
    value=[0] * 1000
)

assessor.launch_profiling()
assessor.plot(path="output_directory/")
```

With "output_directory/" being the output directory of your choice
"""

__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__version__ = "0.0.3"
__date__ = "22/03/2024"
__copyright__ = "MIT License"


# [C]
from .class_performance_assessor import PerformanceAssessor
# [P]
from .parse_argument.parse_argument import parse_argument


def main():
    """Main function.
    """
    __argument = parse_argument(version=__version__)

    assessor: PerformanceAssessor = PerformanceAssessor(
        main=__argument.function,
        n_field=__argument.n_field,
        **__argument.argument
    )

    assessor.launch_profiling()
    assessor.plot(path=__argument.output)


if __name__ == "__main__":
    main()



"""Contains a function to parse given arguments.
"""


__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"

# [I]
from importlib.util import module_from_spec, spec_from_file_location
# [O]
from os.path import exists, isdir
# [S]
from sys import modules

# [Y]
from yaml import safe_load


def check_argument(argument: object) -> object:
    """Check if the different given arguments are good or not.

    Parameters
    ----------
    argument : `ArgumentParser`
        The parsed arguments to checked.

    Returns
    -------
    `ArgumentParser`
        The modified and corrected parsed arguments.
    """
    # Check errors linked to given files.
    __file_errors(argument=argument)
    # Check errors linked to module importations.
    argument = __module_importation_error(argument=argument)

    # Parse the ".yml" file.
    if argument.argument is not None:
        # Parsing user file.
        with open(argument.argument, "r", encoding="utf-8") as file:
            argument.argument = safe_load(file)
    else:
        argument.argument = {}

    return argument

# pylint: disable=too-many-branches
# We have to check if different parameters are good or not. Implying a lot of
# branches.
def __file_errors(argument: object):
    """Check errors linked to given files.

    Parameters
    ----------
    argument : `ArgumentParser`
        The parsed argument to check.

    Raises
    ------
    `FileNotFoundError`
        If the script file is not found.

    `ValueError`
        If the script file have not a ".py" extension.

    `FileNotFoundError`
        If the package file is not found.

    `ValueError`
        If the package file is not a "__init__.py" one.

    `FileNotFoundError`
        If the subpackage file is not found.

    `ValueError`
        If the package file is not a "__init__.py" one.

    `ValueError`
        If the package file is not a "__init__.py" one.

    `FileNotFoundError`
        If the YAML file is not found.

    `ValueError`
        If the YAML file have not a ".yml" extension.

    `FileNotFoundError`
        If the directory is not found.

    `ValueError`
        If the path given is not a directory.

    `ValueError`
        If a number of negative file is given.
    """
    # Input script errors.
    if not exists(argument.script):
        raise FileNotFoundError(f"[Err##] In script, file \"{argument.script}\" "
                                "does not exist.")
    if not argument.script.endswith(".py"):
        raise ValueError(f"[Err##] In script, file \"{argument.script}\" have "
                         "not the extension \".py\".")

    # Input package errors.
    if argument.package is not None:
        if not exists(argument.package):
            raise FileNotFoundError("[Err##] In script, file "
                                    f"\"{argument.package}\" does not exist.")
        if not argument.package.endswith("__init__.py"):
            raise ValueError(f"[Err##] In script, file \"{argument.package}\" "
                             "have to be a \"__init__.py\" file.")

    # Input subpackage errors.
    if argument.subpackage is not None:
        if not exists(argument.subpackage):
            raise FileNotFoundError("[Err##] In script, file "
                                    f"\"{argument.subpackage}\" does not "
                                    "exist.")
        if not argument.subpackage.endswith("__init__.py"):
            raise ValueError("[Err##] In script, file "
                             f"\"{argument.subpackage}\" have to be a "
                             "\"__init__.py\" file.")

    # Input YAML errors.
    if argument.argument is not None:
        if not exists(argument.argument):
            raise FileNotFoundError("[Err##] In script, file "
                                    f"\"{argument.argument}\" does not "
                                    "exist.")
        if not argument.argument.endswith(".yml"):
            raise ValueError("[Err##] In script, file "
                             f"\"{argument.argument}\" have to be a "
                             "\".yml\" file.")

    # Deleting potential "/" at directory end.
    if argument.output[-1] == "/":
        argument.output = argument.output[:-1]

    # Output errors.
    if not exists(argument.output):
        raise FileNotFoundError("[Err##] In output, directory "
                                f"\"{argument.output}\" does not exist.")
    if not isdir(argument.output):
        raise ValueError("[Err##] In output, the given path is not a "
                         "directory.")

    if argument.n_field < 0:
        raise ValueError("[Err##] In n_filed, the value "
                         f"\"{argument.n_field}\" should be greater strictly "
                         "to 0.")

# pylint: enable=too-many-branches


def __module_importation_error(argument: object) -> object:
    """Try to import the given file in a module and do some modifications.

    Parameters
    ----------
    argument : `ArgumentParser`
        The parsed argument to check.

    Returns
    -------
    `ArgumentParser`
        The modified and corrected parsed arguments.

    Raises
    ------
    `ValueError`
        If subpackage is filled, package must be given too.

    `ValueError`
        The given function name is not found in the given module.
    """
    # =============
    #
    # SIMPLE SCRIPT
    #
    # =============
    if argument.package is None and argument.subpackage is None:
        # Load the function inside the module.
        module_spec = spec_from_file_location(
            name=argument.function,
            location=argument.script
        )

        # "Prepare" the module.
        module = module_from_spec(spec=module_spec)

        # "Launch" the module inside the python environment.
        module_spec.loader.exec_module(module=module)
    # =======
    #
    # PACKAGE
    #
    # =======
    elif argument.package is not None and argument.subpackage is None:
        # Get the package name.
        package_name: str = argument.package.split("/")[-2]
        # Load the package.
        package_spec = spec_from_file_location(package_name, argument.package)

        # "Prepare" the package.
        package = module_from_spec(package_spec)
        # Add the package to the modules dict.
        modules[package_name] = package

        # "Launch" the package.
        package_spec.loader.exec_module(package)

        # Set the module name.
        module_name: list = []

        for item in argument.script[:-3].split("/")[::-1]:
            module_name += [item]

            if item == package_name:
                break

        module_name: str = ".".join(module_name[::-1])

        # Load the script inside the module.
        module_spec = spec_from_file_location(module_name, argument.script)

        # "Prepare" the script.
        module = module_from_spec(module_spec)
        # Add the script to the modules dict.
        modules[package_name + ".None"] = module

        # "Launch" the script.
        module_spec.loader.exec_module(module)
    # ==========
    #
    # SUBPACKAGE
    #
    # ==========
    elif argument.package is not None and argument.subpackage is not None:
        # Get the package name.
        package_name: str = argument.package.split("/")[-2]
        # Load the package.
        package_spec = spec_from_file_location(package_name, argument.package)

        # "Prepare" the package.
        package = module_from_spec(package_spec)
        # Add the package to the modules dict.
        modules[package_name] = package

        # "Launch" the package.
        package_spec.loader.exec_module(package)

        # Set the module name.
        module_name: list = []

        for item in argument.script[:-3].split("/")[::-1]:
            module_name += [item]

            if item == package_name:
                break

        module_name: str = ".".join(module_name[::-1])

        # Set the subpackage name.
        subpackage_name: list = []

        for item in argument.subpackage[:-3].split("/")[::-1]:
            subpackage_name += [item]

            if item == package_name:
                break

        subpackage_name: str = ".".join(subpackage_name[::-1])

        # Load the script inside the module.
        module_spec = spec_from_file_location(module_name, argument.script)

        # "Prepare" the script.
        module = module_from_spec(module_spec)
        # Add the script to the modules dict.
        modules[subpackage_name] = module

        # "Launch" the script.
        module_spec.loader.exec_module(module)
    else:
        raise ValueError("[Err##] If --subpackage is specified, --package "
                         "must be specified too.")

    # Check if the function exists inside the module/package/subpackage.
    try:
        argument.function = module.__dict__[argument.function]
    except KeyError as error:
        raise ValueError(f"[Err##] \"{argument.function}\" is not present "
                         "inside the given script "
                         f"\"{argument.script}\".") from error

    return argument



r"""Test if "src/perfassess/parse_argument/check_argument.py" is functional.

Usage
-----
When you are in the main directory ("performance_assessor/"), you can launch:

```sh
    # Execute this command line in a conda environment.
    $ pytest
```
"""

__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"

# [D]
from dataclasses import dataclass

# [P]
import pytest

# [S]
from src.perfassess.parse_argument.check_argument import check_argument


@dataclass
class ArgumentParserSimulator:
    """A class to simulate an argument parser.
    """

    def __init__(
        self,
        script: str,
        output: str,
        function: str,
        n_field: int,
        package: str,
        subpackage: str,
        argument: str
    ):
        """Simulate the creation of parsed arguments.

        Parameters
        ----------
        script : `str`
            The script path.

        output : `str`
            The directory to output results.

        function : `str`
            The function to assess.

        n_field : `int`
            The number of field to keep.

        package : `str`
            A package path.

        subpackage : `str`
            A subpackage path.

        argument : `str`
            A YAML file path.
        """
        self.script: str = script
        self.output: str = output
        self.function: str = function
        self.n_field: int = n_field
        self.package: str = package
        self.subpackage: str = subpackage
        self.argument: str = argument

    def redefine_parameter(self, **kwargs):
        """Give multiple parameters to redefine them.

        Raises
        ------
        `KeyError`
            If the given parameter does not exists.
        """
        for key, value in kwargs.items():
            if key == "script":
                self.script = value
            elif key == "output":
                self.output = value
            elif key == "function":
                self.function = value
            elif key == "n_field":
                self.n_field = value
            elif key == "package":
                self.package = value
            elif key == "subpackage":
                self.subpackage = value
            elif key == "argument":
                self.argument = value
            else:
                raise KeyError("[Err##] Wrong key given.")


@pytest.fixture
def __argument() -> dataclass:
    """Return a instantiated object of `PerformanceAssessor` to test it. Use a
    pytest magic method to only do it once.

    Returns
    -------
    `PerformanceAssessor`
        The class to test.
    """
    return ArgumentParserSimulator(
        script="src/perfassess/main.py",
        output="data/",
        function="main",
        n_field=0,
        package="src/perfassess/__init__.py",
        subpackage="src/perfassess/__init__.py",
        argument="data/argument.yml"
    )


@pytest.mark.parametrize(
    "parameter",
    [
        {"script": "//None//none.py"},
        {"output": "//None"},
        {"package": "//None//none__init__.py"},
        {"subpackage": "//None//none__init__.py"},
        {"argument": "//None//None.yml"}
    ]
)
def test_file_not_found(__argument: dataclass, parameter: dict):
    """Test if the file are not found.

    Parameters
    ----------
    __argument : `dataclass`
        The class that simulates input argument.

    parameter : `dict`
        The tested wrong file path.
    """
    # Satanic notation, please forgive me lord~.
    old_parameter = __argument.__dict__[list(parameter.keys())[0]]

    # Add wrong parameter.
    __argument.redefine_parameter(**parameter)

    # Test the error.
    with pytest.raises(FileNotFoundError):
        check_argument(__argument)

    for key in parameter.keys():
        parameter[key] = old_parameter

    # Give back good parameter.
    __argument.redefine_parameter(**parameter)


@pytest.mark.parametrize(
    "parameter",
    [
        {"script": "data/argument.yml"},
        {"output": "data/argument.yml"},
        {"n_field": -1},
        {"package": "data/argument.yml"},
        {"subpackage": "data/argument.yml"},
        {"argument": "src/perfassess/main.py"}
    ]
)
def test_value_error(__argument: dataclass, parameter: dict):
    """Test if error are well thrown when bad values are given.

    Parameters
    ----------
    __argument : `dataclass`
        The class that simulates input argument.

    parameter : `dict`
        The tested wrong values.
    """
    # Satanic notation, please forgive me lord~.
    old_parameter = __argument.__dict__[list(parameter.keys())[0]]

    __argument.redefine_parameter(**parameter)

    # Test the error.
    with pytest.raises(ValueError):
        check_argument(__argument)

    for key in parameter.keys():
        parameter[key] = old_parameter

    # Give back good parameter.
    __argument.redefine_parameter(**parameter)


@pytest.mark.parametrize(
    "parameter",
    [
        {"function": "none"}
    ]
)
def test_module_error(__argument: dataclass, parameter: dict):
    """Test if an error is thrown when a non-existing function is given.

    Parameters
    ----------
    __argument : `dataclass`
        The class that simulates input argument.

    parameter : `dict`
        The tested wrong function name.
    """
    __argument.redefine_parameter(**parameter)

    # Test the error.
    with pytest.raises(ValueError):
        check_argument(__argument)



"""A simple function to test the module.
"""

__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"


def testor(value: list, to_add: int = 1):
    """A function to launch the program in order to test it.

    Parameters
    ----------
    value : `list`
        A list of values.

    to_add : `int`, optional
        An optional argument to test them. By default 1.
    """
    for value_i in value:
        to_add += value_i



r"""Test if "src/main.py" is functional.

Usage
-----
When you are in the main directory ("performance_assessor/"), you can launch:

```sh
    # Execute this command line in a conda environment.
    $ pytest
```
"""

__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"

# [P]
import pytest

# [P]
from src.perfassess.main import PerformanceAssessor


# =========================================
#
#   Testing the PerformanceAssessor class.
#
# =========================================


@pytest.fixture
def __assessor() -> PerformanceAssessor:
    """Return a instantiated object of `PerformanceAssessor` to test it. Use a
    pytest magic method to only do it once.

    Returns
    -------
    `PerformanceAssessor`
        The class to test.
    """
    performance_assessor: PerformanceAssessor = PerformanceAssessor(
        main=lambda: 10,
        n_field=1
    )

    performance_assessor.launch_profiling()

    return performance_assessor


def test_plot_path_not_exists(__assessor: PerformanceAssessor):
    """Test if an error is thrown when a path that does not exist is given.

    Parameters
    ----------
    __assessor : `PerformanceAssessor`
        The class to test.
    """
    with pytest.raises(FileNotFoundError):
        __assessor.plot("//not_a_directory")


def test_plot_is_not_directory(__assessor: PerformanceAssessor):
    """Test if an error is thrown when a path pointing to something else than a
    file is given.

    Parameters
    ----------
    __assessor : `PerformanceAssessor`
        The class to test.
    """
    with pytest.raises(ValueError):
        __assessor.plot("data/existing_file.txt")


def test_launch_no_profiling(__assessor: PerformanceAssessor):
    """Test if an error is thrown when either `do_memory` and `do_time` are set
    to `False`.

    Parameters
    ----------
    __assessor : `PerformanceAssessor`
        The class to test.
    """
    with pytest.raises(ValueError):
        __assessor.launch_profiling(do_memory=False, do_time=False)



"""Contains a function to parse given arguments.
"""


__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"


# [T]
from textwrap import dedent

# [A]
from argparse import ArgumentParser, RawTextHelpFormatter


def define_argument(version: str = None) -> ArgumentParser:
    """Parse user given arguments.

    Parameters
    ----------
    version : `str`, optional
        The script version. By default None.

    Returns
    -------
    `ArgumentParser`
        The object with unchecked parsed arguments.
    """
    logo: str = """
                ▄▀▀█▄   ▄▀▀▀▀▄  ▄▀▀▀▀▄  ▄▀▀█▄▄▄▄  ▄▀▀▀▀▄  ▄▀▀▀▀▄
               ▐ ▄▀ ▀▄ █ █   ▐ █ █   ▐ ▐  ▄▀   ▐ █ █   ▐ █ █   ▐
                 █▄▄▄█    ▀▄      ▀▄     █▄▄▄▄▄     ▀▄      ▀▄  
                ▄▀   █ ▀▄   █  ▀▄   █    █    ▌  ▀▄   █  ▀▄   █ 
               █   ▄▀   █▀▀▀    █▀▀▀    ▄▀▄▄▄▄    █▀▀▀    █▀▀▀ ▄
               ▐   ▐    ▐       ▐       █    ▐    ▐       ▐     
                                        ▐                        
                     _____________________________________
                    |""|""|""|""|""|""|""|""|""|""|""|""|"|
                    |  1  2  3  4  5  6  7  8  9 10 11 12 |
                    '-------------------------------------'
    """

    print(logo[1:])

    # Description of the program given when the help is cast.
    description: str = """
    "Assess." is a program to assess the time execution of a given script in
    python. To use it, launch:

    \033[7m [[NORMAL USE]] \033[0m\n

        $ perfassess -s script.py \\
                     -f function_name \\
                     -o output_directory/ \\
                     -a argument.yml

    You can actually directly test the program in the repository root
    (perfassess/) using:

        $ perfassess -s src/perfassess/testor.py \\
                     -f testor \\
                     -a data/argument.yml \\
                     -o data/

    \033[7m [[PACKAGE USE]] \033[0m\n
    Let us say that you want to test a package, which should have this kind of
    tree structure:

        package/
        └── src/
            ├── __init__.py
            └── script.py

    In package, you can use relative paths. But with these, the "normal use"
    method do not work. Instead, use the next command, if you are in package/,
    in order to evaluate script.py:

        $ perfassess -s src/script.py \\
                     -f function_name \\
                     -o output_directory/ \\
                     --package src/__init__.py \\
                     -a argument.yml

    In addition of the previous "normal use" method, you have to indicate a
    __init__.py file.

    \033[7m [[SUBACKAGE USE]] \033[0m\n
    Let us say that you want to test a subpackage, which should have this kind
    of tree structure:

        ./package/
        └── src/
            ├── __init__.py
            └── subpackage/
                ├── __init__.py
                └── script.py

    In subpackage, you can use relative paths. But with these, the "normal use"
    method do not work. But the "package use" also do not work. Instead, use
    the next command, if you are in package/, in order to evaluate script.py:

        $ perfassess -s src/subpackage/script.py \\
                     -f function_name \\
                     -o output_directory/ \\
                     --package src/__init__.py \\
                     --subpackage src/subpackage/__init__.py \\
                     -a argument.yml

    In addition of the previous "normal use" and "package use" method, you have
    to indicate two __init__.py files. The first one is the top package
    __init__.py file. The second one is the __init__.py file of the subpackage
    to test.
    """

    # ===================
    #
    # ADD ARGUMENT PARSER
    #
    # ===================

    # Setup the arguments parser object.
    parser: object = ArgumentParser(
        description=dedent(description)[1:-1],
        formatter_class=RawTextHelpFormatter,
        add_help=False
    )

    # == REQUIRED.
    parser.add_argument(
        "-s",
        "--script",
        dest="script",
        required=True,
        type=str,
        metavar="[FILE][\".py\"]",
        help=("\033[7m [[MANDATORY]] \033[0m\n    > The \"scrit.py\" file to "
              "assess.")
    )

    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        required=True,
        type=str,
        metavar="[DIRECTORY]",
        help=("\033[7m [[MANDATORY]] \033[0m\n    > A folder where the "
              "plots will be stored.")
    )

    # == OPTIONAL.
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        help="    > Display this help message, then exit the program."
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"Program version is {version}",
        help="    > Display the program's version, then exit the\nprogram."
    )

    parser.add_argument(
        "-f",
        "--function",
        dest="function",
        required=False,
        default="main",
        type=str,
        metavar="[str|main]",
        help=("    > The entry point to assess a script. It have to be\nthe "
              "name of a function. By default, \"main\".")
    )

    parser.add_argument(
        "--n_field",
        dest="n_field",
        required=False,
        default=0,
        type=int,
        metavar="[int|0]",
        help=("    > The number of field to keep in function name. "
              "By\ndefault 0.")
    )

    parser.add_argument(
        "--package",
        dest="package",
        required=False,
        default=None,
        type=str,
        metavar="[FILE][\".py\"]",
        help="    > The path to an \"__init__.py\" package. By default\nNone."
    )

    parser.add_argument(
        "--subpackage",
        dest="subpackage",
        required=False,
        default=None,
        type=str,
        metavar="[FILE][\".py\"]",
        help=("    > The path to an \"__init__.py\" subpackage. By\ndefault "
              "None.")
    )

    parser.add_argument(
        "-a",
        "--argument",
        dest="argument",
        required=False,
        default=None,
        type=None,
        metavar="[FILE][\".yml\"]",
        help=("    > A YAML file containing all argument for the\nfunction to "
              "test. By default None.")
    )

    argument: ArgumentParser = parser.parse_args()

    return argument



"""Contains a function to parse given arguments.
"""


__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"


# [C]
from .check_argument import check_argument
# [D]
from .define_argument import define_argument


def parse_argument(version: str = None) -> object:
    """Parse given arguments and tests them.

    Parameters
    ----------
    version : `str`, optional
        The script version. By default None.

    Returns
    -------
    `ArgumentParser`
        The object with checked parsed arguments.
    """
    # Parse the arguments.
    argument = define_argument(version=version)
    # Test the arguments.
    argument = check_argument(argument)

    return argument


if __name__ == "__main__":
    __argument = parse_argument()

    print(f"{__argument.function=}")
    print(f"{__argument.script=}")
    print(f"{__argument.output=}")
    print(f"{__argument.argument=}")
    print(f"{__argument.package=}")
    print(f"{__argument.subpackage=}")



r"""An object to compute time and memory consumption.
"""

__authors__ = ["Lucas ROUAUD"]
__contact__ = ["lucas.rouaud@gmail.com"]
__copyright__ = "MIT License"

# [C]
from cProfile import Profile
# [I]
from io import StringIO
# [P]
from pstats import Stats
import sys
# [T]
import tracemalloc
from typing import Callable
# [O]
from os.path import exists, isdir

# [N]
import numpy as np
# [P]
import plotly.graph_objects as go

# [T]
from .testor import testor


class PerformanceAssessor:
    """A class to access the performance of a given function (memory or time).
    """

    def __init__(
        self,
        main: Callable,
        n_field: int = 0,
        **kwargs
    ):
        """Initialize a PerformanceAssessor object.

        Parameters
        ----------
        main : `Callable`
            The function to access.

        n_field : `int`, optional
            The number of field to keep in function name. By default 9.

            Let us say that we have a function named like this:

            ```sh
            /home/user/Document/some_function.py:000(function_name)
            ```

            If input we 2, we will have something like this:

            ```sh
            Document/some_function.py:000(function_name)
            ```

            We split at each "/" and keep the last two field.

        kwargs
            All possible arguments for the function to test.

        Example
        -------
        ```py
        assessor: PerformanceAssessor = PerformanceAssessor(
            main=function_to_test,
            n_field=1,
            argument_for_function_to_test="some_value"
        )
        ```
        """
        self.__assessed_function: Callable = main
        self.__function_argument: dict = dict(kwargs)
        self.__data: dict = {}
        self.__plot: "dict[go.Figure]" = {}
        self.__n_field: int = n_field

    def launch_profiling(
        self,
        do_memory: bool = True,
        do_time: bool = True
    ):
        """Launch the evaluation of performance (memory or time).

        Parameters
        ----------
        do_memory : `bool`, optional
            Do the memory evaluation. By default True.

        do_time : `bool`, optional
            Do the time evaluation. By default True.

        Raises
        ------
        `ValueError`
            When both `do_memory` and `do_time` are set to `False`.
        """
        if not do_memory and not do_time:
            raise ValueError("[Err##] One value between \"do_memory\" or "
                             "\"do_time\" have to set to `True`.")

        if do_memory:
            # Starting to check memory usage.
            tracemalloc.start()

        if do_time:
            profile: Profile = Profile()
            # Starting to check time usage.
            profile.enable()

        # Launch the function to test.
        self.__assessed_function(**self.__function_argument)

        if do_memory:
            # Get a traceback of memory usage.
            snapshot = tracemalloc.take_snapshot()

            # Create the plot for evaluating memory usage.
            self.__memory_evaluation(stat_memory=snapshot.statistics("lineno"))

            # Stop to check memory usage.
            tracemalloc.stop()

        if do_time:
            # Stop to check time usage.
            profile.disable()

            # Create the plot for evaluating memory usage.
            self.__time_evaluation(profile=profile)

    def __memory_evaluation(
        self,
        stat_memory: tracemalloc.Statistic
    ):
        """Parsed memory evaluation output and set a plot.

        Parameters
        ----------
        stat_memory : `tracemalloc.Statistic`
            The "assessor".
        """
        stat_dict = {}

        # Setting the dataset for memory usage.
        for stat in stat_memory:
            key: str = str(stat).split()[0]

            if self.__n_field > 0:
                key = key.split(sep="/")
                key = key[len(key) - self.__n_field:]
                key = "/".join(key)

            if key not in stat_dict:
                stat_dict[key] = 0

            stat_dict[key] += stat.size

        # Save data into member.
        self.__data["memory_evaluation"] = {
            "head": np.array(["size (Mib)", "function"]),
            "label": np.array(list(stat_dict.keys())),
            "data": np.array([list(stat_dict.values())]).T / 1024
        }

        # "Pre-draw" the plot for time usage.
        self.__plot["memory_evaluation"] = self.__set_plot(
            head=np.array(["size (Mib)", "function"]),
            label=np.array(list(stat_dict.keys())),
            data=np.array([list(stat_dict.values())]).T / 1024
        )

    def __time_evaluation(
        self,
        profile: Profile
    ):
        """Parsed time evaluation output and set a plot.

        Parameters
        ----------
        profile : `Profile`
            The "assessor".
        """
        # Redirect print stdout into a buffer.
        buffer: StringIO = StringIO()
        sys.stdout = buffer

        # Get the traceback of time execution.
        stat_time: Stats = Stats(profile).strip_dirs().print_stats()
        stat_time = buffer.getvalue()

        # Set the print stdout standard behaviour.
        sys.stdout = sys.__stdout__

        skip_line: bool = True
        data_label: np.array = np.array([])
        numeric_data: np.array = np.array([])

        # Setting the dataset for time usage.
        for line in stat_time.strip().split("\n"):
            if "ncalls" in line:
                data_head: np.array = np.array(line.split())
                # Add units (seconds) to each header, but the first one.
                data_head = np.char.add(data_head, ["", *[" (s)"] * 4, ""])

                # pylint: disable=C0103
                # It is not a constant pylint!
                skip_line = False
                # pylint: enable=C0103
                continue

            if skip_line or "/" in line:
                continue

            line: list = line.split(maxsplit=5)

            data_label = np.append(data_label, line[-1])

            if numeric_data.shape[0] == 0:
                numeric_data = np.array(line[:-1])
            else:
                numeric_data = np.vstack((numeric_data, line[:-1]))

        numeric_data = numeric_data.astype(float)

        # Save data into member.
        self.__data["time_evaluation"] = {
            "head": data_head,
            "label": data_label,
            "data": numeric_data
        }

        # "Pre-draw" the plot for time usage.
        self.__plot["time_evaluation"] = self.__set_plot(
            head=data_head,
            label=data_label,
            data=numeric_data
        )

    # pylint: disable=too-many-arguments, too-many-instance-attributes
    # Special class that needs a lot of data to be set up. Using a dataclass
    # would be unnecessary, as far as it would just simply take more memory
    # for nothing.

    def __set_plot(
        self,
        head: np.array,
        label: np.array,
        data: np.array,
        foreground: str = "#2E2E3E",
        background: str = "rgba(0, 0, 0, 0)"
    ) -> go.Figure:
        """Set a Plotly bar plot based on computed evaluation.

        Parameters
        ----------
        head : `np.array`
            The data header (like ncall). Or like one label per column.

        label : `np.array`
            The data label (like functions names). Or like one label per row.

        data : `np.array`
            The numerical data.

        foreground : `str`, optional
            The "foreground" color. By default "#2E2E3E".

        background : `str`, optional
            The "background" color. By default "rgba(0, 0, 0, 0)".

        Returns
        -------
        go.Figure
            The setted Plotly bar plot.
        """
        plot: object = go.Figure()

        sort_i: np.array = np.flip(np.argsort(data.T[0]))

        # Trace the barplot
        plot.add_trace(go.Bar(
            x=label[sort_i],
            y=data.T[0][sort_i],
            marker_line_color=foreground,
            marker_color=foreground
        ))

        # Modify general plot properties.
        plot.update_layout(
            template="plotly_white",
            margin={"r": 5},
            font={"size": 12, "family": "Roboto Light"},
            xaxis={
                "title": f"<b>Tested function ({head[-1]})</b>",
                "showline": True,
                "linewidth": 1,
                "showgrid": False,
                "title_font": {"family": "Roboto Black"},
                "tickfont": {"size": 12}
            },
            yaxis={
                "title": f"<b>{head[0].capitalize()}</b>",
                "showline": True,
                "linewidth": 1,
                "showgrid": False,
                "title_font": {"family": "Roboto Black"},
                "tickfont": {"size": 12}
            },
            title_font={"family": "Roboto Black"},
            plot_bgcolor=background,
            paper_bgcolor=background,
        )

        # Add the rectangle border.
        plot.add_shape(
            type="rect",
            xref="paper",
            yref="paper",
            x0=0,
            y0=0,
            x1=1,
            y1=1,
            line={"width": 2, "color": foreground}
        )

        plot.update_traces()

        # Adding the dropdown menu in the case of multiple datas.
        if data.T.shape[0] != 1:
            # Add the dropdown to the plot.
            plot.update_layout(updatemenus=self.__add_dropdown(
                foreground=foreground,
                head=head,
                label=label,
                data=data
            ))

        return plot

    # pylint: enable=too-many-arguments

    def __add_dropdown(
        self,
        head: np.array,
        label: np.array,
        data: np.array,
        foreground: str
    ) -> list:
        """Add a dropdown to the Plotly plot, in order to select different
        assessed values.

        Parameters
        ----------
        head : `np.array`
            The data header (like ncall). Or like one label per column.

        label : `np.array`
            The data label (like functions names). Or like one label per row.

        data : `np.array`
            The numerical data.

        foreground : `str`
            The "foreground" color.

        Returns
        -------
        `list`
            The dropdown menu, which is a `update_menu`.
        """
        button: list = []

        for i, label_i in enumerate(head[:-1]):
            sort_i: np.array = np.flip(np.argsort(data.T[i]))

            # Add a element in the dropdown. By selecting it, it will modify
            # the plot.
            button += [{
                "method": "update",
                "label": label_i,
                "args": [
                    # Restyling.
                    {"x": [label[sort_i]], "y": [data.T[i][sort_i]]},
                    # Updating.
                    {"yaxis": {
                        "showline": True,
                        "linewidth": 1,
                        "showgrid": False,
                        "title": {
                            "text": f"<b>{label_i.capitalize()}</b>",
                            "font": {"family": "Roboto Black"}
                        },
                        "tickfont": {"size": 12}
                    }}
                ],
            }]

        # Create the dropdown menu.
        update_menu: list = [{
            "buttons": button,
            "type": "dropdown",
            "direction": "down",
            "showactive": True,
            "x": 1,
            "xanchor": "right",
            "y": 1.01,
            "yanchor": "bottom",
            "bgcolor": "#FFF",
            "bordercolor": foreground,
            "borderwidth": 2,
            "font_color": foreground
        }]

        return update_menu

    def plot(
        self,
        path: str = "./"
    ):
        """Save the plot to a `.html` file.

        Parameters
        ----------
        path : `str`, optional
            The path to save the file, which have to be a directory. By default
            "./".

        Raises
        ------
        `FileNotFoundError`
            If the input path does not exist.

        `ValueError`
            If the input path is not a directory.
        """
        if path.endswith("/"):
            path = path[:-1]

        if not exists(path):
            raise FileNotFoundError(f"[Err##] Given path \"{path}\" does not "
                                    "exist.")
        if not isdir(path):
            raise ValueError(f"[Err##] Given path  \"{path}\" is not "
                             "directory.")

        for key, plot_i in self.__plot.items():
            # Save the plot.
            plot_i.write_html(
                file=f"{path}/{key}.html",
                include_plotlyjs=True,
                full_html=True
            )

    def get_plot(self) -> dict:
        """Get the setted plot.

        Returns
        -------
        `dict`
            The setted plot.
        """
        return self.__plot

    def data(self) -> dict:
        """Get computed data.

        Returns
        -------
        `dict`
            Computed data about time or memory usage.

        Raises
        ------
        `ValueError`
            _description_
        """
        if not self.__data:
            raise ValueError("[Err##] You have to computed properties before"
                             "getting them. For this, use"
                             "`self.launch_profiling()`.")

        return self.__data


if __name__ == "__main__":
    assessor: PerformanceAssessor = PerformanceAssessor(
        main=testor,
        n_field=1,
        value=[0] * 1000
    )

    assessor.launch_profiling()
    assessor.plot(path="data/")
